<?php

/*
 * Copyright (C) 2016 Champs-Libres <info@champs-libres.coop>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

namespace Chill\ActivityBundle\Export\Filter;

use Chill\MainBundle\Export\FilterInterface;
use Doctrine\ORM\QueryBuilder;
use Symfony\Component\Form\FormBuilderInterface;
use Symfony\Bridge\Doctrine\Form\Type\EntityType;
use Chill\ActivityBundle\Entity\ActivityReason;
use Chill\MainBundle\Templating\TranslatableStringHelper;
use Doctrine\ORM\Query\Expr;
use Symfony\Component\Security\Core\Role\Role;
use Chill\ActivityBundle\Security\Authorization\ActivityStatsVoter;
use Doctrine\ORM\EntityRepository;
use Doctrine\ORM\Query\Expr\Join;
use Symfony\Component\Validator\Context\ExecutionContextInterface;
use Chill\MainBundle\Export\ExportElementValidatedInterface;

/**
 * 
 *
 * @author Julien Fastré <julien.fastre@champs-libres.coop>
 */
class ActivityReasonFilter implements FilterInterface,
   ExportElementValidatedInterface
{
    /**
     *
     * @var TranslatableStringHelper
     */
    protected $translatableStringHelper;
    
    /**
     * The repository for activity reasons
     *
     * @var EntityRepository
     */
    protected $reasonRepository;
    
    public function __construct(
            TranslatableStringHelper $helper,
            EntityRepository $reasonRepository
    ) {
        $this->translatableStringHelper = $helper;
        $this->reasonRepository         = $reasonRepository;
    }
    
    
    public function alterQuery(QueryBuilder $qb, $data)
    {
        $where = $qb->getDQLPart('where');
        $join  = $qb->getDQLPart('join');
        $clause = $qb->expr()->in('reasons', ':selected_activity_reasons');
        //dump($join);
        // add a join to reasons only if needed
        if (
            (array_key_exists('activity', $join)
            &&
            !$this->checkJoinAlreadyDefined($join['activity'], 'reasons')
            )
            OR
            (! array_key_exists('activity', $join))
        ) {
            $qb->add(
                    'join', 
                    array('activity' => new Join(Join::INNER_JOIN, 'activity.reasons', 'reasons')), 
                    true
                    );
        }

        if ($where instanceof Expr\Andx) {
            $where->add($clause);
        } else {
            $where = $qb->expr()->andX($clause);
        }
        
        $qb->add('where', $where);
        $qb->setParameter('selected_activity_reasons', $data['reasons']);
    }
    
    /**
     * Check if a join between Activity and Reason is already defined
     * 
     * @param Join[] $joins
     * @return boolean
     */
    private function checkJoinAlreadyDefined(array $joins, $alias)
    {
        foreach ($joins as $join) {
            if ($join->getAlias() === $alias) {
                return true;
            }
        }
        
        return false;
    }

    public function applyOn()
    {
        return 'activity';
    }

    public function buildForm(FormBuilderInterface $builder)
    {
        //create a local copy of translatableStringHelper
        $helper = $this->translatableStringHelper;
        
        $builder->add('reasons', EntityType::class, array(
            'class' => 'ChillActivityBundle:ActivityReason',
            'choice_label' => function (ActivityReason $reason) use ($helper) {
                return $helper->localize($reason->getName());
            },
            'group_by' => function(ActivityReason $reason) use ($helper) {
                return $helper->localize($reason->getCategory()->getName());
            },
            'multiple' => true,
            'expanded' => false
        ));
    }
    
    public function validateForm($data, ExecutionContextInterface $context)
    {
        if ($data['reasons'] === null || count($data['reasons']) === 0) {
            $context->buildViolation("At least one reason must be choosen")
                ->addViolation();
        }
    }

        public function getTitle()
    {
        return 'Filter by reason';
    }
    
    public function addRole()
    {
        return new Role(ActivityStatsVoter::STATS);
    }
    
    public function describeAction($data, $format = 'string')
    {
        // collect all the reasons'name used in this filter in one array
        $reasonsNames = array_map(
                function(ActivityReason $r) {
                    return "\"".$this->translatableStringHelper->localize($r->getName())."\"";
                },
                $this->reasonRepository->findBy(array('id' => $data['reasons']->toArray()))
                );
        
        return array("Filtered by reasons: only %list%", 
            ["%list%" => implode(", ", $reasonsNames)]);
    }
}
