import { createStore } from 'vuex';
import { datetimeToISO, ISOToDatetime } from 'ChillMainAssets/chill/js/date.js';
import { findSocialActionsBySocialIssue } from 'ChillPersonAssets/vuejs/_api/SocialWorkSocialAction.js';
import { create } from 'ChillPersonAssets/vuejs/_api/AccompanyingCourseWork.js';

const debug = process.env.NODE_ENV !== 'production';

console.log(window.accompanyingCourseWork);

const store = createStore({
  strict: debug,
  state: {
    work: window.accompanyingCourseWork,
    startDate: ISOToDatetime(window.accompanyingCourseWork.startDate.datetime),
    endDate: (window.accompanyingCourseWork.endDate !== null ?
      ISOToDatetime(window.accompanyingCourseWork.endDate.datetime) : null),
    note: window.accompanyingCourseWork.note,
    goalsPicked: window.accompanyingCourseWork.goals,
    resultsPicked: window.accompanyingCourseWork.results,
    resultsForAction: [],
    goalsForAction: [],
    resultsForGoal: [],
    personsPicked: window.accompanyingCourseWork.persons,
    personsReachables: window.accompanyingCourseWork.accompanyingPeriod.participations.filter(p => p.endDate == null)
      .map(p => p.person),
    handlingThirdParty: window.accompanyingCourseWork.handlingThierParty,
    thirdParties: window.accompanyingCourseWork.thirdParties,
    isPosting: false,
    errors: [],
  },
  getters: {
    socialAction(state) {
      return state.work.socialAction;
    },
    hasResultsForAction(state) {
      return state.resultsForAction.length > 0;
    },
    resultsForGoal: (state) => (goal) => {
      let founds = state.resultsForGoal.filter(r => r.goalId === goal.id);

      return founds === undefined ? [] : founds;
    },
    resultsPickedForGoal: (state) => (goal) => {
      let found = state.goalsPicked.find(g => g.goal.id === goal.id);

      return found === undefined ? [] : found.results;
    },
    hasHandlingThirdParty(state) {
      return state.handlingThirdParty !== null;
    },
    hasThirdParties(state) {
      return state.thirdParties.length > 0;
    },
    buildPayload(state) {
      return {
        type: 'accompanying_period_work',
        id: state.work.id,
        startDate: {
          datetime: datetimeToISO(state.startDate)
        },
        endDate: state.endDate === null ? null : {
          datetime: datetimeToISO(state.endDate)
        },
        note: state.note,
        persons: state.personsPicked.map(p => ({id: p.id, type: p.type})),
        handlingThierParty: state.handlingThirdParty === null ? null : {
          id: state.handlingThirdParty.id,
          type: state.handlingThirdParty.type
        },
        results: state.resultsPicked.map(r => ({id: r.id, type: r.type})),
        thirdParties: state.thirdParties.map(t => ({id: t.id, type: t.type})),
        goals: state.goalsPicked.map(g => {
          let o = {
            type: g.type,
            note: g.note,
            goal: {
              type: g.goal.type,
              id: g.goal.id,
            },
            results: g.results.map(r => ({id: r.id, type: r.type})),
          };

          if (g.id !== undefined) {
            o.id = g.id;
          }

          return o;
        })
      };
    }
  },
  mutations: {
    setStartDate(state, date) {
      state.startDate = date;
    },
    setEndDate(state, date) {
      state.endDate = date;
    },
    setResultsForAction(state, results) {
      console.log('set results for action', results);
      state.resultsForAction = results;
    },
    setResultsForGoal(state, { goal, results }) {
      console.log('set results for goal', results);
      state.goalsForAction.push(goal);
      for (let i in results) {
        let r = results[i];
        r.goalId = goal.id;
        console.log('adding result', r);
        state.resultsForGoal.push(r);
      }
    },
    addResultPicked(state, result) {
      state.resultsPicked.push(result);
    },
    removeResultPicked(state, result) {
      state.resultsPicked = state.resultsPicked.filter(r => r.id !== result.id);
    },
    addGoal(state, goal) {
      let g = {
        type: "accompanying_period_work_goal",
        goal: goal,
        note: '',
        results: []
      }
      state.goalsPicked.push(g);
    },
    removeGoal(state, goal) {
      state.goalsPicked = state.goalsPicked.filter(g => g.id !== goal.id);
    },
    addResultForGoalPicked(state, { goal, result}) {
      let found = state.goalsPicked.find(g => g.goal.id === goal.id);
      console.log('adResultForGoalPicked');
      console.log('found', found);
      console.log('goal', goal);
      console.log('result', result);

      if (found === undefined) {
        return;
      }

      found.results.push(result);
    },
    removeResultForGoalPicked(state, { goal, result}) {
      let found = state.goalsPicked.find(g => g.goal.id === goal.id);

      if (found === undefined) {
        return;
      }

      found.results = found.results.filter(r => r.id !== result.id);
    },
    setPersonsPickedIds(state, ids) {
      console.log('persons ids', ids);
      state.personsPicked = state.personsReachables
        .filter(p => ids.includes(p.id))
    },
    setNote(state, note) {
      state.note = note;
    },
    setHandlingThirdParty(state, thirdParty) {
      state.handlingThirdParty = thirdParty;
    },
    addThirdParties(state, thirdParties) {
      console.log('addThirdParties', thirdParties);
      // filter to remove existing thirdparties
      let ids = state.thirdParties.map(t => t.id);
      let unexistings = thirdParties.filter(t => !ids.includes(t.id));
      console.log('unexisting third parties', unexistings);
      for (let i in unexistings) {
        state.thirdParties.push(unexistings[i]);
      }
    },
    removeThirdParty(state, thirdParty) {
      state.thirdParties = state.thirdParties
        .filter(t => t.id !== thirdParty.id);
    },
    setErrors(state, errors) {
      console.log('handling errors', errors);
      state.errors = errors;
    },
    setIsPosting(state, st) {
      state.isPosting = st;
    },
  },
  actions: {
    getReachablesGoalsForAction({ getters, commit, dispatch }) {
      console.log('getReachablesGoalsForAction');
      let
        socialActionId = getters.socialAction.id,
        url = `/api/1.0/person/social-work/goal/by-social-action/${socialActionId}.json`
      ;

      console.log(url);

      window
        .fetch(
          url
        ).then( response => {
          if (response.ok) {
            return response.json();
          }
          throw { m: 'Error while retriving goal for social action', s: response.status, b: response.body };
        }).then( data => {
           for (let i in data.results) {
             dispatch('getReachablesResultsForGoal', data.results[i]);
           }
        }).catch( errors => {
            commit('addErrors', errors);
        });
    },
    getReachablesResultsForGoal({ commit }, goal) {
      console.log('getReachablesResultsForGoal');
      let
        url = `/api/1.0/person/social-work/result/by-goal/${goal.id}.json`
      ;

      console.log(url);

      window.fetch(url)
        .then(response => {
          if (response.ok) {
            return response.json();
          }

          throw { m: 'Error while retriving results for goal', s: response.status, b: response.body };
        })
        .then(data => {
          console.log('data');
          commit('setResultsForGoal', { goal, results: data.results });
        });
    },
    getReachablesResultsForAction({ getters, commit }) {
      console.log('getReachablesResultsForAction');
      let
        socialActionId = getters.socialAction.id,
        url = `/api/1.0/person/social-work/result/by-social-action/${socialActionId}.json`
      ;

      console.log(url);

      window.fetch(url)
        .then(response => {
          if (response.ok) {
            return response.json();
          }

          throw { m: 'Error while retriving results for social action', s: response.status, b: response.body };
        })
        .then(data => {
          console.log('data retrived', data);
          commit('setResultsForAction', data.results);
        });
    },
    submit({ getters, state, commit }) {
      let
        payload = getters.buildPayload,
        url = `/api/1.0/person/accompanying-course/work/${state.work.id}.json`,
        errors = []
      ;

      console.log('action subitting', payload, url);
      commit('setIsPosting', true);

      window.fetch(url, {
        method: 'PUT',
        headers: {
          'Content-Type': 'application/json'
        },
        body: JSON.stringify(payload)
      }).then(response => {
        if (response.ok || response.status === 422) {
          return response.json().then(data => ({data, status: response.status}));
        }

        throw new Error(response.status);
      }).then(({data, status}) => {
        if (status === 422) {
          for (let i in data.violations) {
            errors.push(data.violations[i].title);
          }
          commit('setErrors', errors);
          commit('setIsPosting', false);
        } else {
          console.info('nothing to do here, bye bye');
          window.location.assign(`/fr/person/accompanying-period/${state.work.accompanyingPeriod.id}/work`);
        }
      }).catch(e => {
        commit('setErrors', [
          'Erreur serveur ou réseau: veuillez ré-essayer. Code erreur: ' + e
        ]);
        commit('setIsPosting', false);
      });
    },
    initAsync({ dispatch }) {
      dispatch('getReachablesResultsForAction');
      dispatch('getReachablesGoalsForAction');
    },
  }
});

store.dispatch('initAsync');

export { store };
