<?php

/*
 * Chill is a software for social workers
 * 
 * Copyright (C) 2014-2015, Champs Libres Cooperative SCRLFS, 
 * <http://www.champs-libres.coop>, <info@champs-libres.coop>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

namespace Chill\PersonBundle\Tests\Entity;

use Chill\PersonBundle\Entity\Household\HouseholdMember;
use Chill\PersonBundle\Entity\Household\Position;
use Chill\PersonBundle\Entity\Household\Household;
use Chill\PersonBundle\Entity\Person;
use Chill\PersonBundle\Entity\AccompanyingPeriod;
use Chill\MainBundle\Entity\Address;
use DateInterval;
use DateTime;
use Generator;

/**
 * Unit tests for the person Entity
 */
class PersonTest extends \PHPUnit\Framework\TestCase
{
    /**
     * Test the creation of an accompanying, its closure and the access to 
     * the current accompaniying period via the getCurrentAccompanyingPeriod
     * function.
     */
    public function testGetCurrentAccompanyingPeriod()
    {
        $d = new \DateTime('yesterday'); 
        $p = new Person();
        $p->addAccompanyingPeriod(new AccompanyingPeriod($d));
        
        $period = $p->getCurrentAccompanyingPeriod();
        
        $this->assertInstanceOf(AccompanyingPeriod::class, $period);
        $this->assertTrue($period->isOpen());
        $this->assertEquals($d, $period->getOpeningDate());
        
        //close and test
        $period->setClosingDate(new \DateTime('tomorrow'));
        
        $shouldBeNull = $p->getCurrentAccompanyingPeriod();
        $this->assertNull($shouldBeNull);
    }
    
    /**
     * Test if the getAccompanyingPeriodsOrdered function return a list of
     * periods ordered ascendency.
     */
    public function testAccompanyingPeriodOrderWithUnorderedAccompanyingPeriod()
    {       
        $d = new \DateTime("2013/2/1");
        $p = new Person();
        $p->addAccompanyingPeriod(new AccompanyingPeriod($d));
        
        $e = new \DateTime("2013/3/1");
        $period = $p->getCurrentAccompanyingPeriod()->setClosingDate($e);
        $p->close($period);
        
        $f = new \DateTime("2013/1/1");
        $p->open(new AccompanyingPeriod($f));
        
        $g = new \DateTime("2013/4/1"); 
        $period = $p->getCurrentAccompanyingPeriod()->setClosingDate($g);
        $p->close($period);
        
        $r = $p->getAccompanyingPeriodsOrdered();
        
        $date = $r[0]->getOpeningDate()->format('Y-m-d');
        
        $this->assertEquals($date, '2013-01-01');
    }
    
    /**
     * Test if the getAccompanyingPeriodsOrdered function, for periods
     * starting at the same time order regarding to the closing date.
     */
    public function testAccompanyingPeriodOrderSameDateOpening() {
        $d = new \DateTime("2013/2/1");
        $p = new Person();
        $p->addAccompanyingPeriod(new AccompanyingPeriod($d));
        
        $g = new \DateTime("2013/4/1"); 
        $period = $p->getCurrentAccompanyingPeriod()->setClosingDate($g);
        $p->close($period);
        
        $f = new \DateTime("2013/2/1");
        $p->open(new AccompanyingPeriod($f));
        
        $e = new \DateTime("2013/3/1");
        $period = $p->getCurrentAccompanyingPeriod()->setClosingDate($e);
        $p->close($period);

        $r = $p->getAccompanyingPeriodsOrdered();
        
        $date = $r[0]->getClosingDate()->format('Y-m-d');
        
        $this->assertEquals($date, '2013-03-01');
    }
    
    /**
     * Test if the function checkAccompanyingPeriodIsNotCovering returns
     * the good constant when two periods are collapsing : a period
     * is covering another one : start_1 < start_2 & end_2 < end_1
     */
    public function testDateCoveringWithCoveringAccompanyingPeriod() {
        $d = new \DateTime("2013/2/1");
        $p = new Person();
        $p->addAccompanyingPeriod(new AccompanyingPeriod($d));
        
        $e = new \DateTime("2013/3/1");
        $period = $p->getCurrentAccompanyingPeriod()->setClosingDate($e);
        $p->close($period);
        
        $f = new \DateTime("2013/1/1");
        $p->open(new AccompanyingPeriod($f));
        
        $g = new \DateTime("2013/4/1"); 
        $period = $p->getCurrentAccompanyingPeriod()->setClosingDate($g);
        $p->close($period);
        
        $r = $p->checkAccompanyingPeriodsAreNotCollapsing();
        
        $this->assertEquals($r['result'], Person::ERROR_PERIODS_ARE_COLLAPSING);
    }
    
    /**
     * Test if the function checkAccompanyingPeriodIsNotCovering returns
     * the good constant when two periods are collapsing : a period is open
     * before an existing period
     */
    public function testNotOpenAFileReOpenedLater() {
        $d = new \DateTime("2013/2/1");
        $p = new Person();
        $p->addAccompanyingPeriod(new AccompanyingPeriod($d));
        
        $e = new \DateTime("2013/3/1");
        $period = $p->getCurrentAccompanyingPeriod()->setClosingDate($e);
        $p->close($period);
        
        $f = new \DateTime("2013/1/1");
        $p->open(new AccompanyingPeriod($f));
        
        $r = $p->checkAccompanyingPeriodsAreNotCollapsing();
        
        $this->assertEquals($r['result'], Person::ERROR_ADDIND_PERIOD_AFTER_AN_OPEN_PERIOD);
    }

    public function dateProvider(): Generator
    {
        yield [(DateTime::createFromFormat('Y-m-d', '2021-01-05'))->settime(0, 0)];
        yield [(DateTime::createFromFormat('Y-m-d', '2021-02-05'))->settime(0, 0)];
        yield [(DateTime::createFromFormat('Y-m-d', '2021-03-05'))->settime(0, 0)];
    }

    /**
     * @dataProvider dateProvider
     */
    public function testGetLastAddress(DateTime $date)
    {
        $p = new Person($date);

        // Make sure that there is no last address.
        $this::assertNull($p->getLastAddress());

        // Take an arbitrary date before the $date in parameter.
        $addressDate = clone $date;

        // 1. Smoke test: Test that the first address added is the last one.
        $address1 = (new Address())->setValidFrom($addressDate->sub(new DateInterval('PT180M')));
        $p->addAddress($address1);

        $this::assertCount(1, $p->getAddresses());
        $this::assertSame($address1, $p->getLastAddress());

        // 2. Add an older address, which should not be the last address.
        $addressDate2 = clone $addressDate;
        $address2 = (new Address())->setValidFrom($addressDate2->sub(new DateInterval('PT30M')));
        $p->addAddress($address2);

        $this::assertCount(2, $p->getAddresses());
        $this::assertSame($address1, $p->getLastAddress());

        // 3. Add a newer address, which should be the last address.
        $addressDate3 = clone $addressDate;
        $address3 = (new Address())->setValidFrom($addressDate3->add(new DateInterval('PT30M')));
        $p->addAddress($address3);

        $this::assertCount(3, $p->getAddresses());
        $this::assertSame($address3, $p->getLastAddress());

        // 4. Get the last address from a specific date.
        $this::assertEquals($address1, $p->getLastAddress($addressDate));
        $this::assertEquals($address2, $p->getLastAddress($addressDate2));
        $this::assertEquals($address3, $p->getLastAddress($addressDate3));
    }

    public function testIsSharingHousehold()
    {
        $person = new Person();
        $household = new Household();
        $positionShare = (new Position())
            ->setShareHousehold(true);
        $positionNotShare = (new Position())
            ->setShareHousehold(false);

        $membership1 = (new HouseholdMember())
            ->setStartDate(new \DateTimeImmutable('10 years ago'))
            ->setEndDate(new \DateTimeImmutable('5 years ago'))
            ->setPerson($person)
            ->setPosition($positionShare)
            ;
        $household->addMember($membership1);

        $membership2 = (new HouseholdMember())
            ->setStartDate(new \DateTimeImmutable('4 years ago'))
            ->setEndDate(new \DateTimeImmutable('2 years ago'))
            ->setPerson($person)
            ->setPosition($positionNotShare)
            ;
        $household->addMember($membership2);

        $this->assertEquals(2, $person->getHouseholdParticipations()
            ->count());

        $this->assertFalse($person->isSharingHousehold());
        $this->assertTrue($person->isSharingHousehold(
            new \DateTimeImmutable('6 years ago')));
        $this->assertFalse($person->isSharingHousehold(
            new \DateTimeImmutable('3 years ago')));
    }

}
